#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class MorphBlender{
		protected string[] names;
		public DAZMorph[] morphs;
		protected float[] limits;
		protected MorphManager mmE;
		protected MorphManager mmO;
		
		public MorphBlender(MorphManager mmE, MorphManager mmO){
			names = new string[3];
			morphs = new DAZMorph[3];
			limits = new float[3];
			this.mmE = mmE;
			this.mmO = mmO;
			for(int i = 0; i < 3; i++){
				names[i] = "";
			}
		}
		
		public void addExpression(string name, DAZMorph morph, float limit){
			if(names[2] != ""){
				morphs[2].SetValue(0f);
			}
			for(int i = 2; i > 0; i--){
				if(names[i-1] == ""){
					continue;
				}
				morphs[i] = morphs[i-1];
				names[i]  = names[i-1];
				limits[i] = limits[i-1];
			}
			morphs[0] = morph;
			names[0]  = name;
			limits[0] = limit;
		}
		
		public void setRandomMorph(int stage){
			string name;
			DAZMorph morph;
			float limit;
			MorphManager mm = (stage < 5) ? mmE : mmO;

			name  = mm.getRandomMorphName();
			morph = mm.ms[name].morph;
			limit = mm.ms[name].limit;
			addExpression(name, morph, limit);
		}
		
		public float getLimit(){
			return limits[0];
		}
		
		public void changeExpression(float strength, float ratio){
			float ratioAlt = 1 - ratio;
			if(names[1] == "" || names[0] == names[1]){
				morphs[0].SetValue(strength);
			}else{
				morphs[0].SetValue(strength * ratio);
				morphs[1].SetValue(strength * ratioAlt);
			}
		}
		
		public void stopExpression(){
			mmE.stopExpression();
			mmO.stopExpression();			
		}

		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
	}	
	
	public class MorphManager{
		protected Atom atom;
		protected List<string> enabledMorphs;
		protected string[] allMorphs;
		protected GenerateDAZMorphsControlUI morphControl;
		public Dictionary<string, MorphSetting> ms;
		protected string morphListTag;
		protected string morphSettingTag;
		protected JSONNode jsnroot;
		
		public MorphManager(){
			jsnroot = new JSONNode();
			ms = new Dictionary<string, MorphSetting>();
		}
		
		public void initialize(Atom atom, JSONNode jsnroot, string morphListTag, string morphSettingTag){
			try{
				this.atom = atom;
				JSONStorable geometry = atom.GetStorableByID("geometry");
				DAZCharacterSelector character = geometry as DAZCharacterSelector;
				morphControl = character.morphsControlUI;
			
				this.morphListTag = morphListTag;
				this.morphSettingTag = morphSettingTag;
				this.jsnroot = jsnroot;
				
				JSONNode jsn = this.jsnroot[this.morphListTag];
				for(int i = 0; i < jsn.Count; i++){
					string name = jsn[i];
					DAZMorph morph = morphControl.GetMorphByDisplayName(name);
					if(name == "" || ms.ContainsKey(name)) continue;
					ms.Add(name, new MorphSetting(name, morph));
				}
				
				repackEnabledMorphs();
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
		
		public void overrideSettingsAll(){
			try{
				JSONNode jsn = jsnroot[morphSettingTag];
				if(jsn == null){
					return;
				}
				int num = jsn.Count;
				for(int i = 0; i < num; i++){
					string name = jsn[i]["Name"];
					ms[name].morph   = morphControl.GetMorphByDisplayName(name);
					ms[name].enabled = jsn[i]["Enabled"].AsBool;
					ms[name].limit   = jsn[i]["Limit"].AsFloat;
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
		
		public void overrideSetting(string name, bool enabled, float limit){
			try{
				ms[name].morph   = morphControl.GetMorphByDisplayName(name);
				ms[name].enabled = enabled;
				ms[name].limit   = limit;
				
				repackEnabledMorphs();
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}

		public void rewriteJSON(){
			try{
				foreach(var kvp in ms){
					jsnroot[morphSettingTag][kvp.Key]["Name"] = ms[kvp.Key].name;
					jsnroot[morphSettingTag][kvp.Key]["Enabled"] = ms[kvp.Key].enabled.ToString();
					jsnroot[morphSettingTag][kvp.Key]["Limit"] = ms[kvp.Key].limit.ToString();
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
				
		public void stopExpression(){
			try{
				for(int i = 0; i < enabledMorphs.Count; i++){
					ms[enabledMorphs[i]].morph.SetValue(0.000001f);
//					debug("wave1: " + ms[enabledMorphs[i]].morph.displayName + ": " + ms[enabledMorphs[i]].morph.morphValue);
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
				
		public void repackEnabledMorphs(){
			try{
				enabledMorphs = new List<string>();
				foreach(var kvp in ms){
					if(ms[kvp.Key].enabled){
						enabledMorphs.Add(kvp.Key);
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
			}
		}
		
		public string getRandomMorphName(){
			try{
				int num = UnityEngine.Random.Range(0, enabledMorphs.Count - 1);
				return ms[enabledMorphs[num]].name;
			}catch(Exception e){
				SuperController.LogMessage("error in MorphManager.initialize:" + e);
				return null;
			}
		}
		
		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
	}
	
	public class MorphSetting{
		public string name;
		public bool enabled;
		public float limit;
		public DAZMorph morph;
		
		public MorphSetting(string name, DAZMorph morph){
			this.name = name;
			this.morph = morph;
			limit = 1.0f;
			enabled = true;
		}
	}
	
	public class Expression
	{
		#region common or UI Properties
		public Atom atom;
		public bool expressionOnGlobal = true;
		public float expressionClosingTimeOffset1;
		public float expressionClosingTimeOffset2;
		public EBATCore core;
		public bool disableLog;
		public bool isExpressionActive;
		#endregion
		
		#region expression parameters
		protected GenerateDAZMorphsControlUI morphControl;
		public string[] targetMorphs;
		public DAZMorph[] Emorph;
		public DAZMorph[] Omorph;
		protected DAZMorph EmorphEyesClosed;
		protected DAZMorph mouthMorph;
		protected float expressionOpeningTime;
		protected float targetExpressionStrength = 0f;
		protected float initialExpressionStrength = 0f;
		protected float currentExpressionStrength = 0f;
		protected float targetEyeClose;
		protected float initialEyeClose;
		protected float currentEyeClose;
		protected int expressionPattern = 0;
		protected int expressionPatternOld = 0;
		protected float[] eyeCloseMax;
		protected float eyeCloseStrength = 0.4f;
		protected List<DAZMorph> othersMorphs;

		public float expressionRandomOffsetMin = 0.5f;
		public float expressionRandomOffsetMax = 1.0f;
		public float expressionStrengthBase = 1.0f;
		
		public string[] emoPattern;
		public string expressionPatternName1 = "";
		public string expressionPatternName2 = "";
		
		protected float targetMouthStrength = 0f;
		protected float initialMouthStrength = 0f;
		protected float currentMouthStrength = 0f;
		protected float mouthDuration = 0f;
		protected float lastOpenMouthTime = 0f;
		protected bool isMouthActive = false;
		
		public MorphManager mmEmotion;
		public MorphManager mmOrgasm;
		public MorphBlender mb;
		#endregion
		
		public Expression(EBATCore core){
			atom = core.atom;
			isExpressionActive = false;
			this.core = core;
		}

		public void loadSettings(){
			try{
				expressionOnGlobal = core.settings["General"]["ExpressionOn"].AsBool;
				expressionClosingTimeOffset1 = core.settings["General"]["ExpressionUpTime"].AsFloat;
				expressionClosingTimeOffset2 = core.settings["General"]["ExpressionDownTime"].AsFloat;
				expressionStrengthBase = core.settings["General"]["ExpressionStrengthBase"].AsFloat;
				expressionRandomOffsetMin = core.settings["General"]["ExpressionRandomOffsetMin"].AsFloat;
				expressionRandomOffsetMax = core.settings["General"]["ExpressionRandomOffsetMax"].AsFloat;
				
				int num;
				JSONStorable geometry = atom.GetStorableByID("geometry");
				DAZCharacterSelector character = geometry as DAZCharacterSelector;
				morphControl = character.morphsControlUI;
				
				EmorphEyesClosed = morphControl.GetMorphByDisplayName("Eyes Closed");
				mouthMorph = morphControl.GetMorphByDisplayName("Mouth Open");
				
				JSONNode jsEmorph = core.settings["MorphsForEmotions"];
				JSONNode jsEyeCloseMax = core.settings["EyeCloseMax"];
				num = jsEmorph.Count;
				Emorph = new DAZMorph[num];
				eyeCloseMax = new float[num];
				for(int i = 0; i < num; i++){
					core.logger("jsEmorph[" + i + "]=" + jsEmorph[i]);
					Emorph[i] = morphControl.GetMorphByDisplayName(jsEmorph[i]);
					eyeCloseMax[i] = jsEyeCloseMax[i].AsFloat;
				}
				
				JSONNode jsOmorph = core.settings["MorphsForOrgasm"];
				num = jsOmorph.Count;
				Omorph = new DAZMorph[num];
				for(int i = 0; i < num; i++){
					core.logger("jsOmorph[" + i + "]=" + jsOmorph[i]);
					Omorph[i] = morphControl.GetMorphByDisplayName(jsOmorph[i]);
				}
				
				emoPattern = new string[3];
				
				mmOrgasm = new MorphManager();
				mmOrgasm.initialize(atom, core.settings, "MorphsForOrgasm", "SettingForOrgasm");
				mmOrgasm.overrideSettingsAll();
				mmOrgasm.rewriteJSON();
				core.saveConfig();
				
				mmEmotion = new MorphManager();
				mmEmotion.initialize(atom, core.settings, "MorphsForEmotions", "SettingForEmotions");
				mmEmotion.overrideSettingsAll();
				mmEmotion.rewriteJSON();
				core.saveConfig();		

				mb = new MorphBlender(mmEmotion, mmOrgasm);
				
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void applySettings(){
			othersMorphs = new List<DAZMorph>();
			
			morphControl.onlyShowActive = true;
			morphControl.showType = GenerateDAZMorphsControlUI.ShowType.Pose;
			List<DAZMorph> morphs = morphControl.GetMorphs();
			for(int i = 0; i < morphs.Count; i++){
				if(morphs[i].resolvedRegionName.StartsWith("Pose/Head") && morphs[i].active){
					othersMorphs.Add(morphs[i]);
				}
			}
			
			for(int i = 0; i < othersMorphs.Count; i++){
				core.logger("Set Default Value " + othersMorphs[i].resolvedDisplayName);
				othersMorphs[i].SetDefaultValue();
			}
		}
		
		public void maintenance(){
			
		}
		
		public void update(float time){
			try{
				if(expressionOnGlobal == false){
					return;
				}
				float delta = time - expressionOpeningTime;
				float randomOffset = UnityEngine.Random.Range(expressionRandomOffsetMin, expressionRandomOffsetMax);

				isExpressionActive = true;
				if(delta < expressionClosingTimeOffset1 + randomOffset){
					return;
				}

				expressionOpeningTime = time;
				mb.setRandomMorph(core.getStage());
				targetExpressionStrength = mb.getLimit() * UnityEngine.Random.Range(0.8f,1.0f);
				if(core.isDuringAfterglow){
					targetExpressionStrength /= 2;					
				}
				if(targetExpressionStrength < currentExpressionStrength){
					targetExpressionStrength = currentExpressionStrength + 0.01f;
				}
				initialExpressionStrength = currentExpressionStrength;
				
				targetEyeClose = eyeCloseStrength * UnityEngine.Random.Range(0.0f,0.6f);
				initialEyeClose = currentEyeClose;				
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void manage(float time){
			try{
				float delta = time - expressionOpeningTime;
				float blendRatio = calcBlendNewRatio(delta);
				if(isExpressionActive){
					calcExpressionStrength(delta);
					if(currentExpressionStrength == 0){
						isExpressionActive = false;
					}

					mb.changeExpression(currentExpressionStrength, blendRatio);
					changeEyeClose(currentEyeClose);
				}
				if(isMouthActive){
					calcMouthStrength(time);
					changeMouthSingle();
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void calcExpressionStrength(float delta){
			try{
				float tmpExpressionClosingTimeOffset2;
				tmpExpressionClosingTimeOffset2 = core.isDuringAfterglow ? expressionClosingTimeOffset2 * 1 : expressionClosingTimeOffset2;
				if(delta < expressionClosingTimeOffset1){
					currentExpressionStrength = (targetExpressionStrength - initialExpressionStrength) / expressionClosingTimeOffset1 * delta + initialExpressionStrength;
					currentEyeClose = (targetEyeClose - initialEyeClose) / expressionClosingTimeOffset1 * delta + initialEyeClose;
				}else if(delta < expressionClosingTimeOffset1 + tmpExpressionClosingTimeOffset2){
					currentExpressionStrength = targetExpressionStrength - (targetExpressionStrength/tmpExpressionClosingTimeOffset2) * (delta - expressionClosingTimeOffset1);
					currentEyeClose = targetEyeClose - (targetEyeClose/tmpExpressionClosingTimeOffset2) * (delta - expressionClosingTimeOffset1);
				}else{
					isExpressionActive = false;
					currentExpressionStrength = 0f;
					currentEyeClose = 0f;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void openMouth(float time, float duration){
			try{
				if(expressionOnGlobal == false){
					return;
				}
				if(time < lastOpenMouthTime + duration){
					return;
				}
				isMouthActive = true;
				lastOpenMouthTime = time;
				mouthDuration = duration;
				initialMouthStrength = currentMouthStrength;
				float baseStrength = (float)core.getStage()/10f + 0.5f;
				if(core.isDuringOrgasm){
					baseStrength /= 2; 
				}
				core.logger("baseStrength: " + baseStrength);
				targetMouthStrength = baseStrength * UnityEngine.Random.Range(0.5f,0.8f);
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void calcMouthStrength(float time){
			try{
				float delta = time - lastOpenMouthTime;
				float threshold1 = mouthDuration/3;
				float threshold2 = mouthDuration*2/3;
				float threshold3 = mouthDuration;
				if(delta < threshold1){
					currentMouthStrength = (targetMouthStrength - initialMouthStrength) / (threshold1) * delta + initialMouthStrength;
				}else if(delta < threshold2){
					currentMouthStrength = targetMouthStrength * UnityEngine.Random.Range(0.8f,1.2f);
				}else if(delta < threshold3){
					currentMouthStrength = targetMouthStrength - targetMouthStrength * (delta - threshold2) / (threshold3 - threshold2);
				}else{
					currentMouthStrength = 0f;
					isMouthActive = false;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void changeMouthSingle(){
			try{
				mouthMorph.SetValue(currentMouthStrength);
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public float calcBlendNewRatio(float delta){
			try{
				float ratio = delta / expressionClosingTimeOffset1;
				if(ratio > 1){
					ratio = 1;
				}
				return ratio;
			}catch (Exception e){
				core.loggerE(e);
				return 0f;
			}
		}

		public void stopExpression(){
			try{
				mb.stopExpression();
				EmorphEyesClosed.SetValue(0f);
				mouthMorph.SetValue(0f);
/*
				for(int i = 0; i < 4; i++){
					Emorph[i].SetValue(0f);					
				}
				for(int i = 5; i < 12; i++){
					Omorph[i - 5].SetValue(0f);					
				}
				*/
				currentExpressionStrength = 0f;
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void changeEyeClose(float strength){
			try{
				EmorphEyesClosed.SetValue(strength);
			}catch (Exception e){
				core.loggerE(e);
			}
		}
	}
}